from typing import Literal
from dataclasses import dataclass, field
import os.path
import argparse
import json
import subprocess
import sys
import re

@dataclass
class Node:
	name: str
	libtype: Literal["root", "kernel", ""]  # "" indicates user space
	value: int = 0
	children: list["Node"] = field(default_factory=list)

	def to_js(self, string_list: dict[str, int]):
		def intern_string(x: str):
			if not x:
				return json.dumps(x)
			if x not in string_list:
				string_list[x] = len(string_list)
			return f"$[{string_list[x]}]"
		children = lambda: "[" + ",".join(child.to_js(string_list) for child in self.children) + "]"
		return f"{{n:{intern_string(self.name)},l:{intern_string(self.libtype)},v:{self.value},c:{children()}}}"

class FlameGraphCLI:
	def __init__(self, args):
		self.args = args
		self.stack = Node("all", "root")
		self.root_nodes = {}

	@staticmethod
	def get_libtype_from_dso(dso):
		"""
		when kernel-debuginfo is installed,
		dso points to /usr/lib/debug/lib/modules/*/vmlinux
		"""
		if dso and (dso == "[kernel.kallsyms]" or dso.endswith("/vmlinux")):
			return "kernel"

		if dso and dso.endswith("/node"):
			return "node"

		if dso and dso.endswith("/medooze-media-server.node"):
			return "code"

		native_libs = ["libdl.so.2", "libstdc++.so.6", "libm.so.6", "libgcc_s.so.1", "libpthread.so.0", "libc.so.6"]

		if dso and any(dso.endswith("/" + x) for x in native_libs):
			return "builtin"

		return ""

	@staticmethod
	def find_or_create_node(node, name, libtype):
		for child in node.children:
			if child.name == name:
				return child

		child = Node(name, libtype)
		node.children.append(child)
		return child

	def get_root_node(self, event):
		tid = event.get("sample", {}).get("tid", 0)
		pid = event.get("sample", {}).get("pid", 0)
		comm = event["comm"]

		# event["dso"] sometimes contains /usr/lib/debug/lib/modules/*/vmlinux
		# for user-space processes; let's use pid for kernel or user-space distinction
		if pid == 0:
			node = Node(comm, "kernel")
			self.stack.children.append(node)
			return node

		key = f"{comm} ({pid})"
		if m := re.fullmatch(r"(\d+) (VIEW|SUB|PUB|REL)(\d+)?", comm):
			port, kind, n = m.groups()
			key = f"[{kind}]"

		node = self.root_nodes.get(key)
		if node is None:
			node = Node(key, "")
			self.stack.children.append(node)
			self.root_nodes[key] = node
		return node

	def process_event(self, event):
		tid = event.get("sample", {}).get("tid", 0)
		node = self.root_nodes.get(tid)
		if node is None:
			self.root_nodes[tid] = node = self.get_root_node(event)

		if "callchain" in event:
			for entry in reversed(event["callchain"]):
				dso = entry.get("dso")
				sym = entry.get("sym")
				offset = entry.get("sym_off")
				dsoname = os.path.basename(dso) if dso else None

				name = (sym or {}).get("name")
				if name is None and dso:
					name = f"[{dsoname}]"
				if name is None:
					name = "[unknown]"

				libtype = self.get_libtype_from_dso(dso)
				node = self.find_or_create_node(node, name, libtype)
		else:
			name = event.get("symbol", "[...]")
			libtype = self.get_libtype_from_dso(event.get("dso"))
			node = self.find_or_create_node(node, name, libtype)
		node.value += 1

	def get_report_header(self):
		if self.args.input == "-":
			# when this script is invoked with "perf script flamegraph",
			# no perf.data is created and we cannot read the header of it
			return ""

		try:
			output = subprocess.check_output(["perf", "report", "--header-only"])
			return output.decode("utf-8")
		except Exception as err:  # pylint: disable=broad-except
			print("Error reading report header: {}".format(err), file=sys.stderr)
			return ""

	def trace_end(self):
		string_list = {}
		stacks_json = self.stack.to_js(string_list)
		string_list = [ x for x, _ in sorted(string_list.items(), key=lambda x: x[1]) ]
		stacks_json = f"($ => ({stacks_json}))({json.dumps(string_list)})"

		if self.args.format == "html":
			report_header = self.get_report_header()
			options = {
				"colorscheme": self.args.colorscheme,
				"context": report_header
			}
			options_json = json.dumps(options)

			with open(os.path.join(os.path.dirname(sys.argv[0]), "flamegraph.template.html")) as f:
				output_str = f.read()
			output_str = output_str.replace("/** @options_json **/", options_json)
			output_str = output_str.replace("/** @flamegraph_json **/", stacks_json)

			output_fn = self.args.output or "perf-flamegraph.html"
		else:
			output_str = stacks_json
			output_fn = self.args.output or "perf-flamegraph.json"

		if output_fn == "-":
			with open(sys.stdout.fileno(), "w", encoding="utf-8", closefd=False) as out:
				out.write(output_str)
		else:
			print("dumping data to {}".format(output_fn))
			try:
				with open(output_fn, "w", encoding="utf-8") as out:
					out.write(output_str)
			except IOError as err:
				print("Error writing output file: {}".format(err), file=sys.stderr)
				sys.exit(1)


if __name__ == "__main__":
	parser = argparse.ArgumentParser(description="Create flame graphs (Millicast version, v1.0).")
	parser.add_argument("-f", "--format",
						default="html", choices=["json", "html"],
						help="output file format")
	parser.add_argument("-o", "--output",
						help="output file name")
	parser.add_argument("--template",
						default="/usr/share/d3-flame-graph/d3-flamegraph-base.html",
						help="path to flame graph HTML template")
	parser.add_argument("--colorscheme",
						default="blue-green",
						help="flame graph color scheme",
						choices=["blue-green", "orange"])
	parser.add_argument("-i", "--input",
						help=argparse.SUPPRESS)

	cli_args = parser.parse_args()
	cli = FlameGraphCLI(cli_args)

	process_event = cli.process_event
	trace_end = cli.trace_end
